{-
The LRCs' Agt specification Haskell data type
-}

module AgtTypes where


data Agt = Agt NonTerminals
	       Productions
	       Attributes
	       Equations

instance Show Agt 
	where
	 show (Agt nts
		   prods
		   attrs
		   eqs) =  lrc_agt_start
			++ lrc_basetypes
		 	++ lrc_non_terminals_start
			++ lrc_pre_defined_non_terminals
			++ show nts
			++ lrc_non_terminals_end
			++ lrc_productions_start
		        ++ lrc_pre_defined_productions
			++ concatMap show prods
			++ lrc_productions_end
			++ lrc_entries_start
			++ lrc_entries_end
			++ lrc_attributes_start
			++ concatMap show attrs
			++ lrc_attributes_end
			++ lrc_equations_start
			++ concatMap show eqs
			++ lrc_equations_end
			++ lrc_agt_end


{-
LRCs' Agt specification start
-}

lrc_agt_start = newLine "MD0 TEXT"

lrc_agt_end   = newLine "MD0_END" 

{-
LRCs' Agt base types, case independent
-}

lrc_basetypes = 
	 newLine "BASE_TYPES_START"
      ++ newLine "BASE_TYPE lrc_attr lrc_ATTR lrc_attr_compare ptrf2 lrc_attr_equal"
      ++ newLine "BASE_TYPE lrc_bool lrc_BOOL lrc_bool_compare boolf2 lrc_bool_equal"
      ++ newLine "BASE_TYPE char lrc_CHAR lrc_char_compare shared_char lrc_char_equal"
      ++ newLine "BASE_TYPE int lrc_INT lrc_int_compare shared_int lrc_int_equal"
      ++ newLine "BASE_TYPE lrc_map lrc_MAP lrc_map_compare mapf2 lrc_map_equal"
      ++ newLine "BASE_TYPE lrc_ptr lrc_PTR lrc_ptr_compare ptrf2 lrc_ptr_equal"
      ++ newLine "BASE_TYPE lrc_real lrc_REAL lrc_real_compare realf2 lrc_real_equal"
      ++ newLine "BASE_TYPE lrc_string lrc_STR lrc_string_compare strf2 lrc_string_equal"
      ++ newLine "BASE_TYPES_END"


{-
LRCs' Agt Non Terminals data type
-}

lrc_non_terminals_start = 
	 newLine "NONTERMINALS_START"

lrc_pre_defined_non_terminals =
	 newLine "NT INT P"
      ++ newLine "PRIM2BASE INT lrc_INT"
      ++ newLine "NT BOOL P"
      ++ newLine "PRIM2BASE BOOL lrc_BOOL"
      ++ newLine "NT CHAR P"
      ++ newLine "PRIM2BASE CHAR lrc_CHAR"
      ++ newLine "NT REAL P"
      ++ newLine "PRIM2BASE REAL lrc_REAL"
      ++ newLine "NT STR P"
      ++ newLine "PRIM2BASE STR lrc_STR"
      ++ newLine "NT TOK P"
      ++ newLine "PRIM2BASE TOK lrc_STR"
      ++ newLine "NT PTR P"
      ++ newLine "PRIM2BASE PTR lrc_PTR"
      ++ newLine "NT MAP P"
      ++ newLine "PRIM2BASE MAP lrc_MAP"
      ++ newLine "NT ATTR P"
      ++ newLine "PRIM2BASE ATTR lrc_ATTR"

data NonTerminals = NT [NonTerminal]
	
instance Show NonTerminals 
	where
	 show (NT (h:t)) 
	  =   space "NT" ++ space h ++ newLine "r"
	   ++ concatMap (("NT "++).newLine) t
			 

type NonTerminal = String

lrc_non_terminals_end   = 
	 newLine "NONTERMINALS_END"


{-
LRCs' Agt Productions data type
-}

lrc_productions_start =
	 newLine "PRODUCTIONS_START"

lrc_pre_defined_productions =
	 newLine "PROD  _Int INT"
      ++ newLine "NOCC INT N"
      ++ newLine "PROD  _Bool BOOL"
      ++ newLine "NOCC BOOL N"
      ++ newLine "PROD  _Char CHAR"
      ++ newLine "NOCC CHAR N"
      ++ newLine "PROD  _Real REAL"
      ++ newLine "NOCC REAL N"
      ++ newLine "PROD  _Str STR"
      ++ newLine "NOCC STR N"
      ++ newLine "PROD  _Tok TOK"
      ++ newLine "NOCC TOK N"
      ++ newLine "PROD  _Ptr PTR"
      ++ newLine "NOCC PTR N"
      ++ newLine "PROD  _Map MAP"
      ++ newLine "NOCC MAP N"
      ++ newLine "PROD  _Attr ATTR"
      ++ newLine "NOCC ATTR N"
	
type Productions = [Production]

{-
Each production consists in the following:
	the constructor
	the production's lhs nonterminal
	the production's list of nonterminals

for example, the Haskell element 
	Prod 
	  Cons
	  List
	  [List, INT, List]

would represent the LRC production:

List : Cons (INT List)
-}

data Production = Prod 
		    Constructor 
		    NonTerminal
		    [NonTerminal]

instance Show Production 
	where
	 show (Prod 
		cons
		nonT
		l) =  space "PROD " 
		 ++ space cons
		 ++ newLine nonT
		 ++ concatMap (newLine.(space "NOCC"++).(++" N")) l

lrc_productions_end =
	 newLine "PRODUCTIONS_END"	


{-
LRC's Agt Entries data type
-}

lrc_entries_start = newLine "ENTRIES_START"
lrc_entries_end   = newLine "ENTRIES_END"



{-
LRCs' Agt Attributes data type
-}

lrc_attributes_start = newLine "ATTRIBUTES_START"

lrc_attributes_end   = newLine "ATTRIBUTES_END"

type Attributes = [Attribute]

data Attribute = Attr AttributeName
		      AttributeType
		      NonTerminal
	deriving Eq

instance Show Attribute
	where
	 show (Attr
		 name
		 typ
		 nt) =  space "ATTR" 
		     ++ space name 
		     ++ space (show typ)
		     ++ newLine nt 

type AttributeName = String
 
data AttributeType = SYN
		   | INH
	deriving (Eq,Show)

{-
LRCs' Agt Equations data type
-}

lrc_equations_start = newLine "EQUATIONS_START"

lrc_equations_end   = newLine "EQUATIONS_END"
	
type Equations = [Equation]

data Equation = Eq EquationType
		   Constructor
		   Attrs

instance Show Equation
	where
	 show (Eq eqType constructor attrs)
	  = newLine ("EQ " ++ show eqType) 
	    ++ aux constructor attrs 
	  where
	  aux :: Constructor -> Attrs -> String
	  aux constructor [] 
	   = ""
	  aux constructor ((t,occurs,attr):tail)
	   = let str = aux constructor tail
	     in  "AO " ++ space (show t) 
	    	       ++ space (constructor)
		       ++ space (show occurs)
		       ++ newLine attr
		       ++ str	 



data EquationType = SynRule
		  | CopyRule
	deriving Eq

instance Show EquationType
	where 
	 show SynRule 
	  = "S"
	 show CopyRule
	  = "C"

type Constructor = String

type Attrs = [(T,Occurs,Attr)]

data T = A
       | N
	deriving (Show, Eq)

type Occurs = Int
type Attr   = String

{-
Auxiliary combinators
-}

newLine x = x ++ "\n"
space   x = x ++ " "
