module Aux where

import AgtTypes

import FunctionTypes

import Language.Haskell.Syntax

import List

{- 
   We use 'eficientPartition' for partitioning strings in two: the non-numeric part and the numeric part!
   The function is more general, though, since it is possible to use another criteria for partitioning 
	strings.

   Our strings correspond to attributes and their numeric parts, if any, at the end, correspond to attribute
	occurrences,

   For example, if we have the following grammar production:

   Tree: Fork (Int Tree Tree)

   Then, when using "ilist2", we are referring to the attribut "ilist" of the first rhs occurrence of "Tree" 

-}

eficientPartition :: (Char -> Bool) 
		  -> [String] 
	          -> [Char]
	          -> (T, Int, [Char])
eficientPartition p varnames attr 
	= if (attr `elem` varnames)
	  then (N, (\(Just x) -> x+1) (elemIndex attr  varnames),"")
	  else 
	    let (str,n) = eficientPartition' p attr
	    in if (n == [])
	       then (A, 0,                           str)
	       else (A, ((read :: String -> Int) n), str)
	where eficientPartition' _ [] = ([],[])
	      eficientPartition' p (h:t) = let (ac1',ac2') = eficientPartition' p t 
					   in if (p h) 
					      then (ac1',h:ac2')
					      else (h:ac1',ac2')




{- 
some auxiliary functions to acess Haskell variable names to avoid writing 
	'(... (HsIdent ...) ...)' everywhere in the code
-}
getVarName       (UnQual 
		    (HsIdent name)) = name
getPatBindLhsIds (HsPVar 
		    (HsIdent x))    = [x]
getPatBindLhsIds (HsPTuple l)       = concatMap getPatBindLhsIds l 


aux (HsInt int)       = show int
aux (HsChar char)     = show char
aux (HsString string) = show string
aux (HsFrac frac)     = show frac
aux' (HsQVarOp name)  = aux'' name
aux' (HsQConOp name)  = aux'' name
aux'' (Qual _ (HsIdent name))  = name
aux'' (Qual _ (HsSymbol name)) = name
aux'' (UnQual (HsIdent name))  = name
aux'' (UnQual (HsSymbol name)) = name
aux'' (Special HsCons)         = ":"

