module ML_Parser where

import Char

import Parser
import Language.Haskell.Syntax
import Language.Haskell.Pretty

breaks = satisfy (\x -> (x == ' ') || (x=='\n') || (x=='\t') || (x=='\r'))

zeroOuMais parser =  succeed [] 
		 <|> umOuMais parser

umOuMais parser = (:[]) <$> parser 
	       <|> (:)  <$> parser <*> umOuMais parser

token' t = sf <$> aux <*> token t <*> aux							
	where sf _ x _ = x
	      aux = zeroOuMais breaks

token'' t = sf <$> zeroOuMaisEspacos <*> token t <*> zeroOuMaisEspacos	
	where sf _ x _ = x

token''' t = sf <$> token t <*> aux
	where sf x _ = x
	      aux = zeroOuMais breaks

zeroOuMaisEspacos = zeroOuMais (satisfy (==' '))
umOuMaisEspacos   = umOuMais   (satisfy (==' '))

parseML = f1 <$> parseDataDeclarations <*> umOuMais breaks
				       <*> parseFunctionDeclarations
	where f1 x _ y = HsModule srcLoc
				  (Module "ML")
				  Nothing
				  []
				  (x++y)

parseDataDeclarations = succeed []
		    <|> f1 <$> token' "type" <*> parseType
					     <*> parseTypes
	where f1 _ x y = x:y

parseTypes = succeed []
	 <|> f1 <$> token' "type" <*> parseType
				  <*> parseTypes
	 <|> f2 <$> token' "and"  <*> parseType
				  <*> parseTypes
	where f1 _ x y = x:y
	      f2 = f1

parseType = f1 <$> parseVarName <*> token' "="
				<*> parseDecl
				<*> parseDecls
	where f1 x _ y z = HsDataDecl srcLoc
				      []
				      (HsIdent x')
				      []
				      (y:z)
				      []
			where x' = (toUpper (head x)):tail x

parseDecls = succeed []
	 <|> f1 <$> token' "|" <*> parseDecl
			       <*> parseDecls
	where f1 _ x y = x:y

parseDecl = f1 <$> parseVarName <*> token' "of"
				<*> parseArg
				<*> parseArgs
	<|> f2 <$> parseVarName
	where f1 x _ y z = HsConDecl srcLoc 
				     (HsIdent x)
				     (aux y z)
	      aux y z    = if (z==[]) then [HsUnBangedTy y]
				      else [HsUnBangedTy (HsTyTuple (y:z))]
	      f2 x       = HsConDecl srcLoc
				     (HsIdent x)
				     [] 

parseArgs = succeed []
	<|> f1 <$> token' "*" <*> parseArg
			      <*> parseArgs
	where f1 _ x y = x:y

parseArg = f1 <$> parseVarName
	where f1 x = HsTyCon $ UnQual $ HsIdent x'
			where x' = (toUpper (head x)):tail x

parseFunctionDeclarations = f1 <$> token' "let" <*> parseFunction
						<*> parseFunctions
			<|> f2 <$> token' "let rec"
						<*> parseFunction
						<*> parseFunctions
			<|> succeed []
	where f1 _ x y   = x:y
	      f2 _ x y = x:y

parseFunctions = f1 <$> token' ";;"  <*> token' "let"
				     <*> parseFunction
				     <*> parseFunctions
	     <|> f2 <$> token' ";;"  <*> token' "let rec"
				     <*> parseFunction
				     <*> parseFunctions
	     <|> f3 <$> token' "and" <*> parseFunction
				     <*> parseFunctions
	     <|> f4 <$> token' ";;"
	where f1 _ _ x y   = x:y
	      f2 _ _ x y = x:y
	      f3 _ x y     = x:y
	      f4 _         = []

parseFunction = f1 <$> parseVarName <*> umOuMais (satisfy (==' '))
				    <*> parseFunctionArgs
				    <*> token'' "="
				    <*> parseFunctionBody
	where f1 x _ (HsPVar (HsIdent y)) _ (z,w) | y == z = HsFunBind $ map (\(a,b,c) -> HsMatch srcLoc
												(HsIdent x)
												[a]
												b
												c) w
						  | otherwise = error "match mismatch!"
	      f1 x _ (HsPTuple 
			((HsPVar 
				(HsIdent y)):t)) _ (z,w)
						  | y == z = HsFunBind $ map (\(a,b,c) -> HsMatch srcLoc
												(HsIdent x)
												[HsPTuple (a:t)]
												b
												c) w
						  | otherwise = error "match mismatch!"

parseFunctionBody = f1 <$> token' "match" <*> parseVarName
					  <*> token' "with"
					  <*> parseFunctionBodyCases
	where f1 _ x _ y = (x,y)

parseFunctionBodyCases = f1 <$> parseBodyCase <*> parseFunctionBodyCasesAux
	where f1 x y = x:y

parseFunctionBodyCasesAux = succeed []
			<|> f1 <$> token' "|" <*> parseFunctionBodyCases
	where f1 _ x = x

parseBodyCase = f1 <$> parseVarName <*> umOuMais (satisfy (==' '))
				    <*> parseFunctionArgs
				    <*> token' "->"
				    <*> token''' "let"
				    <*> parseLetCase
				    <*> token' "in"
				    <*> parseLetCases
	    <|> f2 <$> parseVarName <*> umOuMais (satisfy (==' '))
				    <*> token' "->"
				    <*> token''' "let"
				    <*> parseLetCase
				    <*> token' "in"
				    <*> parseLetCases
	where f1 x _ y _ _ z _ (w1,w2)
				= (HsPApp (UnQual (HsIdent x)) [y],
				   w1,
				   z:w2
				)
	      f2 x _ _ _ z _ (w1,w2)
				= (HsPApp (UnQual (HsIdent x)) [],
				   w1,
				   z:w2
				)

parseLetCases = f1 <$> token' "let" <*> parseLetCase
				    <*> token' "in"
				    <*> parseLetCases
	    <|> f2 <$> parseFunctionResults
	where f1 _ x _ (y1,y2) = (y1, x:y2)
	      f2 x             = (HsUnGuardedRhs x, [])

parseLetCase = f1 <$> parseFunctionArgs <*> token'' "="
					<*> parseLetCaseAux
	where f1 x _ y = HsPatBind srcLoc x y []

parseLetCaseAux = f1 <$> parseFunctionResults
	      <|> f2 <$> parseVarName <*> zeroOuMaisEspacos
				      <*> parseFunctionResults
	where f1 x     = HsUnGuardedRhs x
	      f2 x _ y | isUpper (head x) = HsUnGuardedRhs (HsApp (HsCon (UnQual (HsIdent x)))
								   y)
		       | otherwise =	    HsUnGuardedRhs (HsApp (HsVar (UnQual (HsIdent x)))
								   y)

parseFunctionArgs = f1 <$> parseVarName
		<|> f2 <$> symbol '(' <*> parseTupleArgs
				      <*> symbol ')'
	where f1 x     = HsPVar (HsIdent x)
	      f2 _ x _ | length x == 1 = head x
		       | otherwise     = HsPTuple x

parseFunctionResults = f1 <$> parseVarName 
		   <|> f2 <$> symbol '(' <*> parseTupleArgs'
				         <*> symbol ')'
	where f1 x     | x == "[]" = HsList []
		       | isInt x   = HsLit (HsInt (read x))
		       | otherwise = HsVar (UnQual (HsIdent x))
		where isInt = foldr (\x y -> (isDigit x)&&y) True 
	      f2 _ x _ | length x == 1 = head x
		       | otherwise     = HsTuple x

parseTupleArgs = f1 <$> zeroOuMaisEspacos <*> parseVarName <*> zeroOuMaisEspacos
	     <|> f2 <$> zeroOuMaisEspacos <*> parseVarName
					  <*> zeroOuMaisEspacos 
				          <*> symbol ','
				          <*> parseTupleArgs
	     
	where f1 _ x _         = [HsPVar (HsIdent x)]
	      f2 _ x _ _ y     = (HsPVar (HsIdent x)):y
							   

parseTupleArgs' = f1 <$> zeroOuMaisEspacos <*> parseVarName <*> zeroOuMaisEspacos
	      <|> f3 <$> zeroOuMaisEspacos <*> parseVarName
					   <*> token'' "(" 
					   <*> parseTupleArgs'
					   <*> token'' ")"
	      <|> f2 <$> zeroOuMaisEspacos <*> parseVarName
					   <*> token'' "," 
				           <*> parseTupleArgs'
	      <|> f4 <$> zeroOuMaisEspacos <*> parseVarName
					   <*> token'' "(" 
					   <*> parseTupleArgs'
					   <*> token'' ")"
					   <*> token' ","
					   <*> parseTupleArgs'	      
	where f1 _ x _     = [HsVar (UnQual (HsIdent x))]
	      f2 _ x _ y   | isUpper (head x) = [HsCon (UnQual (HsIdent x))]++y
			   | otherwise        = [HsVar (UnQual (HsIdent x))]++y
	      f3 _ x _ y _ | isUpper (head x) = if (length y == 1) then [(HsApp (HsCon (UnQual (HsIdent x)))
						                                (head y))]
								   else [(HsApp (HsCon (UnQual (HsIdent x)))
									        (HsTuple y))]
			   | otherwise =        if (length y == 1) then [(HsApp (HsVar (UnQual (HsIdent x)))
									        (head y))]
								   else [(HsApp (HsVar (UnQual (HsIdent x)))
									        (HsTuple y))]
	      f4 _ x _ y _ _ z 
			   | isUpper (head x) = if (length y == 1) then [(HsApp (HsCon (UnQual (HsIdent x)))
						                                (head y))]++z
								   else [(HsApp (HsCon (UnQual (HsIdent x)))
									        (HsTuple y))]++z
			   | otherwise =        if (length y == 1) then [(HsApp (HsVar (UnQual (HsIdent x)))
									        (head y))]++z
								   else [(HsApp (HsVar (UnQual (HsIdent x)))
									        (HsTuple y))]++z


parseVarName = f <$> umOuMais (satisfy (\x -> isAlphaNum x || x == '_' || x == '[' || x == ']'))
	where f str = str


srcLoc = SrcLoc {srcFilename="x",srcLine=0,srcColumn=0}
