module OcamlFunctions where

import Language.Haskell.Syntax

import FunctionTypes
import Aux

import List
import Char

patbinds2OCamlFunctions :: [HsDecl] 
	       -> [String]
	       -> [(String,String,[String])]
patbinds2OCamlFunctions [] 
	       _ = []


-- attr1 = (expr)
patbinds2OCamlFunctions ((HsPatBind x1
		           attr
			   (HsUnGuardedRhs (HsParen expr))
			   x2):t)
		varnames
	= patbinds2OCamlFunctions ((HsPatBind x1 attr (HsUnGuardedRhs expr) x2):t) varnames


-- attr1 = literal
patbinds2OCamlFunctions ((HsPatBind _
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsLit lit))
			  _):t) 
		varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      function       = ( show (i1+1)++"_"++str1,
	        		 aux lit,
				 []
				)	      		       
	  in function : patbinds2OCamlFunctions t varnames 


-- attr1 = function literal
patbinds2OCamlFunctions ((HsPatBind _ 
			  (HsPVar (HsIdent attr1)) 
			  (HsUnGuardedRhs (HsApp (HsVar (UnQual (HsIdent function))) (HsLit lit))) 
			  _):t)
		varnames 
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      eq             = ( show (i1+1)++"_"++str1,
		 		 function ++ " " ++ aux lit,
				 []
				 )
	  in eq : patbinds2OCamlFunctions t varnames 


-- attr1 = []
-- attr1 = [attr2]
patbinds2OCamlFunctions ((HsPatBind _
			    (HsPVar (HsIdent attr1))
			    (HsUnGuardedRhs (HsList expr2))
			    _):t) 
		varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      r              = patbinds2OCamlFunctions t varnames
	  in case expr2 of [] -> 
				let function 
					= (show (i1+1) ++"_" ++str1,
				   	   "[]",
				   	   []
				   	  )
				in function:r
			   [HsVar (UnQual (HsIdent attr2))] 
			      -> 
				let a2@(_,i2,str2) = eficientPartition isDigit varnames attr2
			    	    function 
					= (show (i1+1)++"_"++str1,
				           "(function x -> [x])",
				   	   [show i2 ++ "_" ++ str2])
			        in function:r


-- attr1 = attr2
patbinds2OCamlFunctions ((HsPatBind _
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsVar attr2))
			  _):t) 
		varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      function       = (show (i1+1) ++ "_" ++ str1,
				 "",
				 [show i2 ++ "_" ++ str2]) 
	      in function : patbinds2OCamlFunctions t varnames 


-- attr1 = Constructor
-- where Constructor has 0 arguments
-- falta testar s o construtor est definido no mdulo!
-- s estiver, isto est correcto,seno,  s tirar o "C_" e o "_1"
patbinds2OCamlFunctions ((HsPatBind _
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsCon cons))
			   _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	  in (show (i1+1) ++ "_" ++ str1,
	      "C_" ++ getVarName cons ++ "_1",
	      [])
	      : patbinds2OCamlFunctions t varnames


-- attr1 = Constructor (expr1)
patbinds2OCamlFunctions ((HsPatBind x
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsApp (HsCon cons) (HsParen expr1)))
			   y):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind x
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsApp (HsCon cons) expr1))
			             y):t) varnames


-- attr1 = Constructor literal
patbinds2OCamlFunctions ((HsPatBind _
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsApp (HsCon cons)
						  (HsLit lit)))
			   _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	  in (show (i1+1) ++ "_" ++ str1,
	        "C_"++getVarName cons++"_1 " ++ aux lit,
		[])
	     : patbinds2OCamlFunctions t varnames


-- attr1 = Constructor attr2
patbinds2OCamlFunctions ((HsPatBind _
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsApp (HsCon cons)
						  (HsVar attr2)))
			   _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	  in (show (i1+1) ++ "_" ++ str1,
	        "C_"++getVarName cons++"_1",
		[show i2 ++ "_" ++ str2])
	     : patbinds2OCamlFunctions t varnames


-- attr1 = Constructor (expr2) expr3
patbinds2OCamlFunctions ((HsPatBind b
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsApp (HsApp (HsCon cons)
							 (HsParen expr2))
						  expr3))
			   x):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind b
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsApp (HsApp (HsCon cons)
							           expr2)
							    expr3))
				     x):t) varnames


-- attr1 = Constructor expr2 (expr3)
patbinds2OCamlFunctions ((HsPatBind b
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsApp (HsApp (HsCon cons)
							 expr2)
						  (HsParen expr3)))
			   x):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind b
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsApp (HsApp (HsCon cons)
							           expr2)
							    expr3))
				     x):t) varnames


-- attr1 = Constructor (attr1, attr2)
patbinds2OCamlFunctions ((HsPatBind _
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsCon cons)
						 (HsTuple [expr2, expr3])))
			   _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      r              = patbinds2OCamlFunctions t varnames
	  in case expr2 of
		HsVar attr2 -> case expr3 of
					HsVar attr3 -> -- Constructor (attr2, attr3)
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	  						in  (show (i1+1) ++ "_" ++ str1,
		 					     "(function x -> function y -> C_" ++ aux'' cons ++ "_1 (x, y)) ",
		 					       [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3]
								)
	     						     : r


-- attr1 = Constructor/special_function attr2              attr3
-- attr1 = Constructor/special_function literal            attr2
-- attr1 = Constructor/special_function [function literal] attr2
-- attr1 = Constructor/special_function attr2              literal
-- attr1 = Constructor/special_function [Constructor attr2] attr3
-- attr1 = Constructor/special_function [Constructor attr2 attr3] attr4
patbinds2OCamlFunctions ((HsPatBind _
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsApp (HsApp (HsCon cons)
							 expr2)
						  expr3))
			   _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
              r              = patbinds2OCamlFunctions t varnames
	  in case expr2 of
		HsVar attr2 -> case expr3 of
					HsVar attr3 -> -- attr1 = Constructor attr2 atrr3
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	  						in (show (i1+1) ++ "_" ++ str1,
		 					       "C_" ++ aux'' cons ++ "_1",
		 					       [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3]
								)
	     						    : r
					HsLit lit   -> -- attr1 = Constructor attr2 literal
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
							in (show (i1+1) ++ "_"++ str1,
							     "(function x -> " ++ "C_" ++ aux'' cons ++ "_1 x " ++ aux lit ++ ")",
							     [show i2 ++ "_" ++ str2] 
							    )
	     						    : r
		HsLit lit  -> case expr3 of
					HsVar attr2 -> -- attr1 = Constructor literal attr2
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	  						in (show (i1+1) ++ "_"++ str1,
							      "C_" ++ aux'' cons ++ "_1 " ++ aux lit,
							      [show i2 ++ "_" ++ str2]
							     )
	     						    : r
		HsApp (HsVar f)
		      (HsLit lit) -> case expr3 of
					HsVar attr2 -> -- attr1 = Constructor [function literal] attr2
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	  						in (show (i1+1) ++ "_"++ str1,
		 					      "C_" ++ aux'' cons ++ "_1 (" ++ FunctionTypes.space (getVarName f) ++  aux lit ++ ")",
		 					      [show i2 ++ "_" ++ str2]
							     )
	     						    : r
		HsApp (HsCon (UnQual (HsIdent constructor)))
		      (HsVar attr2)
				-> case expr3 of
					HsVar attr3 -> -- attr1 = Constructor/special_function [Constructor attr2] attr3
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	  						in (show (i1+1) ++ "_" ++ str1,
		 					     "(function x y -> (" ++ aux'' cons ++ ") (C_" ++ constructor ++ "_1  x) y)",
							     [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3]
							   )
	     						    : r
		HsApp (HsApp (HsCon (UnQual (HsIdent constructor)))
			     (HsVar attr2))
		      (HsVar attr3)
				-> case expr3 of 
					HsVar attr4 -> -- attr1 = Constructor/special_function [Constructor attr2 attr3] attr4
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      						    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
	  						in (show (i1+1) ++ "_" ++ str1,
		 					      "(function x y z -> (" ++ aux'' cons ++ ") (C_" ++ constructor ++ "_1  x y) z)",
							      [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4]
							     )
	    	 					    : r


-- attr1 = Constructor (attr2, attr3, attr4)
patbinds2OCamlFunctions ((HsPatBind _
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsCon cons)
						 (HsTuple [expr2, expr3, expr4])))
			   _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      r              = patbinds2OCamlFunctions t varnames
	  in case expr2 of
		HsVar attr2 -> case expr3 of
					HsVar attr3 -> case expr4 of
								HsVar attr4 -> -- Constructor (attr2, attr3, attr4)
									let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    		    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      						    		    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
	  								in (show (i1+1) ++ "_" ++ str1,
		 					       			"(function x -> function y -> function z -> C_" ++ aux'' cons ++ "_1 (x,y,z))",
		 					       			[show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4]
									     )
	     						   		    : r


-- attr1 = Constructor (expr2) expr3 expr4
patbinds2OCamlFunctions ((HsPatBind b
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsCon cons) 
								(HsParen expr2))
							 expr3) 
						  expr4))
			   x):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind b
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsCon cons)
									  expr2)
								   expr3)
						      expr4))
				     x):t) varnames


-- attr1 = Constructor expr2 (expr3) expr4
patbinds2OCamlFunctions ((HsPatBind b
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsCon cons) 
								expr2)
							 (HsParen expr3)) 
						  expr4))
			   x):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind b
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsCon cons)
									  expr2)
								   expr3)
						      expr4))
				     x):t) varnames


-- attr1 = Constructor expr2 expr3 (expr4)
patbinds2OCamlFunctions ((HsPatBind b
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsCon cons) 
								expr2)
							 expr3) 
						  (HsParen expr4)))
			   x):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind b
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsCon cons)
									  expr2)
								   expr3)
						      expr4))
				     x):t) varnames


-- attr1 = Constructor/special_function attr2 attr3 attr4
-- attr1 = Constructor/special_function literal attr3 attr4
-- attr1 = Constructor/special_function [function literal] attr3 attr4
-- attr1 = Constructor/special_function [function attr2] attr3 attr4, 
patbinds2OCamlFunctions ((HsPatBind _
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsCon cons) 
								expr2)
							 expr3) 
						  expr4))
			   _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      r              = patbinds2OCamlFunctions t varnames
	  in case expr2 of 
			HsVar attr2 -> case expr3 of
						HsVar attr3 -> case expr4 of
									HsVar attr4 -> let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
											   a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
											   a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
										       in  (show (i1+1) ++ "_" ++ str1,
		 										"C_" ++ aux'' cons ++ "_1",
		 										[show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4]
											     )
	      										     : r
			HsLit lit   -> case expr3 of
						HsVar attr2 -> case expr4 of
									HsVar attr3 -> let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      										   a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
										       in (show (i1+1) ++ "_"++ str1,
		 										"C_" ++ aux'' cons ++ "_1 " ++ aux lit,
		 										[show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3]
											    )
	     										   : r
			HsApp (HsVar f) (HsLit lit)
				    -> case expr3 of 
						HsVar attr2 -> case expr4 of
									HsVar attr3 -> let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      										   a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
										       in (show (i1+1) ++ "_"++ str1,
		 									     "C_" ++ aux'' cons ++ "_1 (" ++ FunctionTypes.space (getVarName f) ++  aux lit ++ ")",
		 									     [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3]
											     )
	     										     : r
			HsApp (HsVar f) (HsVar attr2)
				   -> case expr3 of
						HsVar attr3 -> case expr4 of
									HsVar attr4 -> let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      									           a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      									           a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
								 		       in (show (i1+1) ++ "_"++ str1,
		 									     "(function x -> function y -> function z -> " ++ "C_" ++ aux'' cons ++ "_1 (" ++ FunctionTypes.space (getVarName f) ++  "x) y z)",
		 									     [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4]
											    )
	     										   : r


-- attr1 = function (expr2)
patbinds2OCamlFunctions ((HsPatBind x
			  (HsPVar attr1)
			  (HsUnGuardedRhs (HsApp (HsVar f)
						 (HsParen expr2)))
			  y):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind x
					(HsPVar attr1)
					(HsUnGuardedRhs (HsApp (HsVar f)
							       expr2))
			 		y):t) varnames


-- attr1 = function attr2
patbinds2OCamlFunctions ((HsPatBind _
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsVar (UnQual (HsIdent f)))
						 (HsVar (UnQual (HsIdent attr2)))))
			  _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      a2@(_,i2,str2) = eficientPartition isDigit varnames attr2
	  in (show (i1+1) ++ "_" ++ str1,
		     f,
		     [show i2 ++ "_" ++ str2]
		    )
		  : patbinds2OCamlFunctions t varnames


-- attr1 = function (expr2) expr3
patbinds2OCamlFunctions ((HsPatBind x
			  (HsPVar attr1)
			  (HsUnGuardedRhs (HsApp (HsApp (HsVar (UnQual f))
							(HsParen expr2))
						 expr3))
			  y):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind x
			  	(HsPVar attr1)
			  	(HsUnGuardedRhs (HsApp (HsApp (HsVar (UnQual f))
						              expr2)
						       expr3))
			  y):t) varnames


-- attr1 = function expr2 (expr3)
patbinds2OCamlFunctions ((HsPatBind x
			  (HsPVar attr1)
			  (HsUnGuardedRhs (HsApp (HsApp (HsVar (UnQual f))
							expr2)
						 (HsParen expr3)))
			  y):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind x
			  	(HsPVar attr1)
			  	(HsUnGuardedRhs (HsApp (HsApp (HsVar (UnQual f))
						              expr2)
						       expr3))
			  y):t) varnames


-- attr1 = function attr2 atrr3
-- attr1 = function Constructor attr2
-- attr1 = function [Constructor attr2] attr3
patbinds2OCamlFunctions ((HsPatBind _
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsApp (HsVar (UnQual (HsIdent f)))
							expr2)
						 expr3))
			  _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      r = patbinds2OCamlFunctions t varnames
	  in case expr2 of
		HsVar attr2 -> case expr3 of 
					HsVar attr3 -> -- attr1 = function  attr2 attr3
						       let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						   a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	  					       in (show (i1+1) ++ "_" ++ str1,
		     					     f,
		     					     [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3]
		    					    )
		  					  : r 
		HsCon cons -> case expr3 of 
					HsVar attr2 -> -- attr1 = function Constructor attr2
						       let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
						       in (show (i1+1) ++ "_" ++ str1,
							      f ++ " " ++ aux'' cons,
							      [show i2 ++ "_" ++ str2]
							     )
							   : r
		HsApp (HsCon cons)
		      (HsVar attr2) -> case expr3 of 
					HsVar attr3 -> -- attr1 = function [Constructor attr2] attr3
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
							    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
							in  (show (i1+1) ++ "_" ++ str1,
							       "(function x -> function y -> " ++ f ++ " (" ++ aux'' cons ++ " x) y)",
							      [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3])
							     : r


-- attr1 = function (expr2) expr3 expr4
patbinds2OCamlFunctions ((HsPatBind x
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsVar (UnQual (HsIdent f)))
							       (HsParen expr2))
							expr3)
						 expr4))
			  y):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind x
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsVar (UnQual (HsIdent f)))
							       expr2)
							expr3)
						 expr4))
			  y):t) varnames


-- attr1 = function expr2 (expr3) expr4
patbinds2OCamlFunctions ((HsPatBind x
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsVar (UnQual (HsIdent f)))
							       expr2)
							(HsParen expr3))
						 expr4))
			  y):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind x
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsVar (UnQual (HsIdent f)))
							       expr2)
							expr3)
						 expr4))
			  y):t) varnames


-- attr1 = function expr2 expr3 (expr4)
patbinds2OCamlFunctions ((HsPatBind x
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsVar (UnQual (HsIdent f)))
							       expr2)
							expr3)
						 (HsParen expr4)))
			  y):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind x
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsVar (UnQual (HsIdent f)))
							       expr2)
							expr3)
						 expr4))
			  y):t) varnames


-- attr1 = function attr2            attr3 attr4
-- attr1 = function (function attr2) attr3 attr4
patbinds2OCamlFunctions ((HsPatBind _
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsVar (UnQual (HsIdent f)))
							       expr2)
							expr3)
						 expr4))
			  _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      r              = patbinds2OCamlFunctions t varnames
	  in case expr2 of
			HsVar attr2 -> case expr3 of -- attr1 = function attr2 attr3 attr4
					HsVar attr3 -> case expr4 of
								HsVar attr4 -> let
	      								a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      								a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      								a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
	  								in (show (i1+1) ++ "_" ++ str1,
		     								f,
		    								[show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4]):r
			(HsApp (HsVar (UnQual (HsIdent f2)))
			       (HsVar attr2))
				    -> case expr3 of -- attr1 = function (function attr2) attr3 attr4
					HsVar attr3 -> case expr4 of
								HsVar attr4 -> let
									a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      								a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      								a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
	  								in (show (i1+1) ++ "_" ++ str1,
		     								"(function x -> function y -> function z -> " ++ f ++ " (" ++ f2 ++" x) y z)",
		    								[show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4]):r


-- attr1 = function (literal, attr2)
-- attr1 = function (attr2, attr3)
-- attr1 = function (attr2, attr3, attr4)
-- attr1 = function (attr2, attr3, attr4, attr5)
-- attr1 = function (attr2, attr3, attr4, attr5, attr6)
-- attr1 = function (attr2, attr3, attr4, attr5, attr6, attr7)
-- attr1 = function (Constructor (attr2, attr3), attr4)
patbinds2OCamlFunctions ((HsPatBind _
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsApp (HsVar (UnQual (HsIdent f)))
						  (HsTuple l)))
			   _):t) varnames 
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      r              = patbinds2OCamlFunctions t varnames
	  in  case l of
		[HsLit lit, HsVar attr2]
					   -> -- attr1 = function (literal, attr2)
			let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
			in  (show (i1+1) ++ "_" ++ str1,
				"(function x -> " ++ f ++ " (" ++ aux lit ++ ", x))",
				[show i2 ++ "_" ++ str2])
			     : r
		[HsApp (HsCon (UnQual (HsIdent cons)))
		       (HsTuple [HsVar attr2, HsVar attr3]), HsVar attr4]
					   -> -- attr1 = function (Constructor (attr2, attr3), attr4)
			let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      		    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      		    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
			in (show (i1+1) ++ "_" ++ str1,
 			      "(function x -> function y -> function z -> " ++ f ++ " (" ++ cons ++ "(x,y), z))",
			      [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4]
			     )
			    : r
		[HsVar attr2, HsVar attr3] 
					   -> -- attr1 = function (attr2, attr3)
			let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      		    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
			in ( show (i1+1) ++ "_" ++ str1,
			      "(function x -> function y -> " ++ f ++ "(x,y))",
			      [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3]
				)
			    : r
		[HsVar attr2, HsVar attr3, HsVar attr4] 
					   -> -- attr1 = function (attr2, attr3, attr4)
			let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      		    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      		    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
			in ( show (i1+1) ++ "_" ++ str1,
				"(function x -> function y -> function z -> " ++ f ++ "(x,y,z))",
				[show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3, show i4 ++ "_" ++ str4]
				)
			    : r
		[HsVar attr2, HsVar attr3, HsVar attr4, HsVar attr5] 
					   -> -- attr1 = function (attr2, attr3, attr4, attr5)
			let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      		    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      		    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
	      		    a5@(_,i5,str5) = eficientPartition isDigit varnames (getVarName attr5)
			in ( show (i1+1) ++ "_" ++ str1,
				"(function x -> function y -> function z -> function w" ++ f ++ "(x,y,z,w))",
				[show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3, show i4 ++ "_" ++ str4, show i5 ++ "_" ++ str5]
				)
			    : r
		[HsVar attr2, HsVar attr3, HsVar attr4, HsVar attr5, HsVar attr6] 
					   -> -- attr1 = function (attr2, attr3, attr4, attr5, attr6)
			let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      		    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      		    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
	      		    a5@(_,i5,str5) = eficientPartition isDigit varnames (getVarName attr5)
	      		    a6@(_,i6,str6) = eficientPartition isDigit varnames (getVarName attr6)
			in ( show (i1+1) ++ "_" ++ str1,
				"(function x -> function y -> function z -> function w -> function u" ++ f ++ "(x,y,z,w,u))",
				[show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3, show i4 ++ "_" ++ str4, show i5 ++ "_" ++ str5, show i6 ++ "_" ++ str6]
				)
			    : r
		[HsVar attr2, HsVar attr3, HsVar attr4, HsVar attr5, HsVar attr6, HsVar attr7] 
					   -> -- attr1 = function (attr2, attr3, attr4, attr5, attr6, attr7)
			let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      		    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      		    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
	      		    a5@(_,i5,str5) = eficientPartition isDigit varnames (getVarName attr5)
	      		    a6@(_,i6,str6) = eficientPartition isDigit varnames (getVarName attr6)
	      		    a7@(_,i7,str7) = eficientPartition isDigit varnames (getVarName attr7)
			in ( show (i1+1) ++ "_" ++ str1,
				"(function x -> function y -> function z -> function w -> function u -> function r" ++ f ++ "(x,y,z,w,u,r))",
				[show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3, show i4 ++ "_" ++ str4, show i5 ++ "_" ++ str5, show i6 ++ "_" ++ str6, show i7 ++ "_" ++ str7]
				)
			    : r

	
-- attr1 = (expr1) `op` expr2
patbinds2OCamlFunctions ((HsPatBind b
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsInfixApp (HsParen expr2)
						       op
						       expr3))
			   x):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind b 
				    (HsPVar (HsIdent attr1))
				    (HsUnGuardedRhs (HsInfixApp expr2
								op
								expr3))
				     x):t) varnames


-- attr1 = expr1 `op` (expr2)
patbinds2OCamlFunctions ((HsPatBind b
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsInfixApp expr2
						       op
						       (HsParen expr3)))
			   x):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind b 
				    (HsPVar (HsIdent attr1))
				    (HsUnGuardedRhs (HsInfixApp expr2
								op
								expr3))
				     x):t) varnames

-- attr1 =  attr2  `op`  attr3
-- attr1 = [attr2  `op`  attr3]  `op`  attr4
-- attr1 =  attr2  `op`  literal
-- attr1 = literal `op`  attr2
-- attr1 = [attr2  `op1` attr3]  `op2` literal
-- attr1 = [function attr2]      `op`  literal
-- attr1 = [function attr2]      `op`  attr3
-- attr1 = [attr2  `op`  attr3]  `op` [attr4   `op` attr5]
-- attr1 = [(function attr2)     `op` literal] `op` attr3
-- attr1 = ((f (Cons attr2)) attr3 attr4) `op` attr5
-- attr1 = (Cons (Cons attr2) attr3) `op` attr4
patbinds2OCamlFunctions ((HsPatBind b
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsInfixApp expr2
						       op
						       expr3))
			   x):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
              r              = patbinds2OCamlFunctions t varnames
	  in case expr2 of
		HsApp (HsCon (UnQual (HsIdent cons)))
		      (HsTuple [HsVar attr2, HsVar attr3])
			-> case expr3 of
					HsVar attr4 ->
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      						    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
							in (show (i1+1) ++ "_" ++ str1,
							     "(function x -> function y -> function z -> (" ++ aux' op ++ ") (" ++ cons ++ " (x,y)) z)",
							     [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3, show i4 ++ "_" ++ str4]
								)
							     : r
		HsApp (HsApp  (HsApp (HsVar f)
			     	     (HsParen (HsApp (HsCon (UnQual (HsIdent cons)))
					     	     (HsVar attr2))))
		              (HsVar attr3))
		      (HsVar  attr4)
			-> case expr3 of 
					HsVar attr5 -> -- attr1 = ((f (Cons attr2)) attr3 attr4) `op` attr5 
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      						    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
	      						    a5@(_,i5,str5) = eficientPartition isDigit varnames (getVarName attr5)
							in ( show (i1+1) ++ "_" ++ str1,
							     "(function x -> function y -> function z -> function w -> (" ++ aux' op ++ ") " ++ "((" ++ getVarName f ++ " (" ++ cons ++ " x)) y z) w)",
							     [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3, show i4 ++ "_" ++ str4, show i5 ++ "_" ++ str5]
	       						     )
							     : r
		HsApp (HsApp (HsCon (UnQual (HsIdent cons1))) 
			     (HsParen (HsApp (HsCon (UnQual (HsIdent cons2)))
			                     (HsVar attr2))))
		      (HsVar attr3)
			-> case expr3 of 
					(HsVar attr4) -> -- attr1 = (Cons (Cons attr2) attr3) `op` attr4
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      						    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
							in (show (i1+1) ++ "_" ++ str1,
								"(function x -> function y -> function z -> (" ++ aux' op ++ ") (" ++ cons1 ++ " (" ++ cons2 ++" x) y) z)",
								[show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3, show i4 ++ "_" ++ str4]
							      ) : r
		HsVar attr2 
			-> case expr3 of
					HsVar attr3 -> -- attr1 = attr2 `op` attr3
						       let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						   a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	  					       in (show (i1+1) ++ "_" ++ str1,
							     "(" ++ aux' op ++ ")",
							    [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3]
	       						    )
	     						   : r
					HsLit lit -> -- attr1 = attr2 `op` literal
						     let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
						     in (show (i1+1) ++ "_" ++ str1,
							   "(" ++ aux' op ++ aux lit ++ ")",
							   [show i2 ++ "_" ++ str2]
	       						  )
	     						 : r
		HsInfixApp (HsVar attr2)
			   op1
			   (HsVar attr3) 
			-> case expr3 of
					HsVar attr4 -> -- attr1 = [attr2 `op` attr3] `op` attr4
						       let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						   a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      						   a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
						       in (show (i1+1) ++ "_" ++ str1,
							    "(function x -> function y -> function z -> ((x " ++ aux' op1 ++ " y) " ++ aux' op ++ " z))",
					 		    [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3, show i4 ++ "_" ++ str4]
       							    )
	     						   : r
					HsLit lit   -> -- attr1 = [attr2 `op1` attr3] `op2` literal
						        let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
							in (show (i1+1) ++ "_" ++ str1,
							      "(function x -> function y -> ((x " ++ aux' op1 ++ " y) " ++ aux' op ++ " " ++ aux lit ++ "))",
							      [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3]
	       						     )
	     						    : r
					HsInfixApp (HsVar attr4)
						   op3
						   (HsVar attr5)
						    -> -- attr1 = [attr2 `op` attr3] `op` [attr4 `op` attr5]
						       let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						   a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      						   a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
						           a5@(_,i5,str5) = eficientPartition isDigit varnames (getVarName attr5)
	  						in (show (i1+1) ++ "_" ++ str1,
							     "(function x -> function y -> function z -> function w-> ((x " ++ aux' op1 ++ " y) " ++ aux' op ++ "(z" ++ aux' op3 ++ "w)))",
							     [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3, show i4 ++ "_" ++ str4, show i5 ++ "_" ++ str5]
	       						     )
	     						    : r
		HsLit lit
			-> case expr3 of 
					HsVar attr3 -> -- attr1 = literal `op` attr2
						       let a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
						       in (show (i1+1) ++ "_" ++ str1,
		 					    aux lit ++ aux' op,
		 					    [show i3 ++ "_" ++ str3]
							    )
	     						   : r
		HsApp  (HsVar f)
		       (HsVar attr2)
			-> case expr3 of
					HsLit lit   -> -- attr1 = [function attr2] `op` literal
						       let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
						       in (show (i1+1) ++ "_" ++ str1,
						 	     "(function x -> (" ++ FunctionTypes.space (getVarName f) ++ "x) " ++ FunctionTypes.space (aux' op) ++ aux lit ++ ")",
							     [show i2 ++ "_" ++ str2])
	     						   : r
					HsVar attr3 -> -- attr1 = [function attr2] `op` attr3
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
							in (show (i1+1) ++ "_" ++ str1,
							     "(function x -> function y -> (" ++ FunctionTypes.space (getVarName f) ++ "x) " ++ FunctionTypes.space (aux' op) ++ "y)",
	       						     [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3]
							     )	
	     						    : r
		HsInfixApp (HsParen (HsApp (HsVar f)
					   (HsVar attr2)))
			   op1
			   (HsLit lit)
			-> case expr3 of 
					HsVar attr3 -> -- attr1 = [(function attr2) `op` literal] `op` attr3
						       let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						   a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
						       in (show (i1+1) ++ "_"++ str1,
		 					     "(function x -> function y -> ((" ++ FunctionTypes.space (getVarName f) ++  "x) " 
			  				     ++ FunctionTypes.space (aux' op1) ++ aux lit ++ ")" ++ aux' op ++ " y)",
		 					     [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3]
	        					    )
	     						  : r
		x -> error (show x)


-- attr1 = if (expr1) then expr2 else expr3
patbinds2OCamlFunctions ((HsPatBind x
			   attr1
			   (HsUnGuardedRhs (HsIf (HsParen expr1) expr2 expr3)) y):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind x attr1 (HsUnGuardedRhs (HsIf expr1 expr2 expr3)) y):t) varnames


-- attr1 = if expr1 then (expr2) else expr3
patbinds2OCamlFunctions ((HsPatBind x
			   attr1
			   (HsUnGuardedRhs (HsIf expr1 
						(HsParen expr2) 
						expr3)) 
			   y):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind x attr1 (HsUnGuardedRhs (HsIf expr1 expr2 expr3)) y):t) varnames


-- attr1 = if expr1 then expr2 else (expr3)
patbinds2OCamlFunctions ((HsPatBind x
			   attr1
			   (HsUnGuardedRhs (HsIf expr1 
						 expr2 
						 (HsParen expr3))) 
			   y):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind x attr1 (HsUnGuardedRhs (HsIf expr1 expr2 expr3)) y):t) varnames


-- attr1 = if [b attr2] then [attr3] else [attr4]
patbinds2OCamlFunctions ((HsPatBind x
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsIf (HsApp (HsVar (UnQual (HsIdent b)))  (HsVar (UnQual (HsIdent attr2))))
						 (HsVar (UnQual (HsIdent attr3)))
						 (HsVar (UnQual (HsIdent attr4)))))
			    y):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind x
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsIf (HsApp (HsVar (UnQual (HsIdent b)))  (HsVar (UnQual (HsIdent attr2))))
						 (HsApp (HsVar (UnQual (HsIdent "id"))) (HsVar (UnQual (HsIdent attr3))))
						 (HsApp (HsVar (UnQual (HsIdent "id"))) (HsVar (UnQual (HsIdent attr4))))))
			   y):t) varnames


-- attr1 = if [b attr2] then [f attr3] else [attr4]
patbinds2OCamlFunctions ((HsPatBind x
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsIf (HsApp (HsVar (UnQual (HsIdent b))) (HsVar (UnQual (HsIdent attr2))))
						 (HsApp (HsVar (UnQual (HsIdent f))) (HsVar (UnQual (HsIdent attr3))))
						 (HsVar (UnQual (HsIdent attr4)))))
			    y):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind x
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsIf (HsApp (HsVar (UnQual (HsIdent b))) (HsVar (UnQual (HsIdent attr2))))
						 (HsApp (HsVar (UnQual (HsIdent f))) (HsVar (UnQual (HsIdent attr3))))
						 (HsApp (HsVar (UnQual (HsIdent "id"))) (HsVar (UnQual (HsIdent attr4))))))
			   y):t) varnames


-- attr1 = if [b attr2] then [attr3] else [f attr4]
patbinds2OCamlFunctions ((HsPatBind x
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsIf (HsApp (HsVar (UnQual (HsIdent b))) (HsVar (UnQual (HsIdent attr2))))
						 (HsVar (UnQual (HsIdent attr3)))
						 (HsApp (HsVar (UnQual (HsIdent f))) (HsVar (UnQual (HsIdent attr4))))))
			    y):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind x
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsIf (HsApp (HsVar (UnQual (HsIdent b))) (HsVar (UnQual (HsIdent attr2))))
						 (HsApp (HsVar (UnQual (HsIdent "id"))) (HsVar (UnQual (HsIdent attr3))))
						 (HsApp (HsVar (UnQual (HsIdent f))) (HsVar (UnQual (HsIdent attr4))))))
			   y):t) varnames


-- attr1 = if [b attr2] then [f1 attr3] else [f2 attr4]
-- assuming that no parenthesis are used to establish priorities
-- the "if ... then ... else ..." is really necessary in order to produce haskell function macros in the way that the cpp accepts them
patbinds2OCamlFunctions ((HsPatBind x
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsIf (HsApp (HsVar (UnQual (HsIdent b)))  (HsVar (UnQual (HsIdent attr2))))
						 (HsApp (HsVar (UnQual (HsIdent f1))) (HsVar (UnQual (HsIdent attr3))))
						 (HsApp (HsVar (UnQual (HsIdent f2))) (HsVar (UnQual (HsIdent attr4))))))
			   _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      a2@(_,i2,str2) = eficientPartition isDigit varnames attr2
	      a3@(_,i3,str3) = eficientPartition isDigit varnames attr3
	      a4@(_,i4,str4) = eficientPartition isDigit varnames attr4
	in (show (i1+1) ++ "_"++ str1,
		 if (attr2==attr3) then "(function x -> function y -> if (" ++ b  ++  " x) then (" ++ f1 ++ " x) else (" ++ f2 ++" y))"
					       else (if (attr2==attr4) 
							then "(function x -> function y -> if (" ++ b  ++  " x) then (" ++ f1 ++ " y) else (" ++ f2 ++ " x))" 
							else (if (attr3==attr4) 
								then "(function x -> function y -> if (" ++ b  ++  " x) then (" ++ f1 ++ " y) else (" ++ f2 ++ " y))"
								else "(function x -> function y -> function z -> if (" ++ b  ++  " x) then (" ++ f1 ++ " y) else (" ++ f2 ++ " z))")),
		 nub [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4]
	        )
	     : patbinds2OCamlFunctions t varnames


-- attr1 = if [attr2 `op` attr3] then attr4 else attr5
-- assuming that no parenthesis are used to establish priorities
patbinds2OCamlFunctions ((HsPatBind x
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsIf (HsInfixApp (HsVar attr2)
							     op
							     (HsVar attr3)) 
						 (HsVar attr4)
						 (HsVar attr5))) y):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind x
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsIf (HsInfixApp (HsVar attr2)
							               op
							               (HsVar attr3))
						            (HsApp (HsVar (UnQual (HsIdent "id"))) (HsVar attr4))
							    (HsApp (HsVar (UnQual (HsIdent "id"))) (HsVar attr5)))) y):t) varnames


-- attr1 = if [attr2 `op` attr3] then [f attr4] else attr5
-- assuming that no parenthesis are used to establish priorities
patbinds2OCamlFunctions ((HsPatBind x
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsIf (HsInfixApp (HsVar attr2)
							     op
							     (HsVar attr3)) 
						 (HsApp (HsVar f) (HsVar attr4))
						 (HsVar attr5))) y):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind x
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsIf (HsInfixApp (HsVar attr2)
							               op
							               (HsVar attr3))
						            (HsApp (HsVar f) (HsVar attr4))
							    (HsApp (HsVar (UnQual (HsIdent "id"))) (HsVar attr5)))) y):t) varnames


-- attr1 = if [attr2 `op` attr3] then [f attr4] else attr5
-- assuming that no parenthesis are used to establish priorities
patbinds2OCamlFunctions ((HsPatBind x
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsIf (HsInfixApp (HsVar attr2)
							     op
							     (HsVar attr3)) 
						 (HsVar attr4)
						 (HsApp (HsVar f) (HsVar attr5)))) y):t) varnames
	= patbinds2OCamlFunctions ((HsPatBind x
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsIf (HsInfixApp (HsVar attr2)
							               op
							               (HsVar attr3))
						            (HsApp (HsVar (UnQual (HsIdent "id"))) (HsVar attr4))
							    (HsApp (HsVar f) (HsVar attr5)))) y):t) varnames


-- attr1 = if [attr2 `op` attr3] then [f1 attr4] else [f2 attr5]
-- the "if ... then ... else ..." chain is really necessary in order to produce haskell function macros 
-- in the way that the cpp accepts them
patbinds2OCamlFunctions ((HsPatBind x
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsIf (HsInfixApp (HsVar (UnQual (HsIdent attr2)))
							     op
							     (HsVar (UnQual (HsIdent attr3))))
						 (HsApp (HsVar (UnQual (HsIdent f1))) (HsVar (UnQual (HsIdent attr4))))
						 (HsApp (HsVar (UnQual (HsIdent f2))) (HsVar (UnQual (HsIdent attr5))))))
			    _):t) varnames 
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      a2@(_,i2,str2) = eficientPartition isDigit varnames attr2
	      a3@(_,i3,str3) = eficientPartition isDigit varnames attr3
	      a4@(_,i4,str4) = eficientPartition isDigit varnames attr4
	      a5@(_,i5,str5) = eficientPartition isDigit varnames attr5
	  in (show (i1+1) ++ "_"++ str1,
		 if (attr2==attr4 && attr3==attr5) then "(function x -> function y -> if (x " ++ (aux' op)  ++  " y) then (" ++ f1 ++ " x) else (" ++ f2 ++ " y))"
					       else (if (attr2==attr5 && attr3==attr4) 
							then "(function x -> function y -> if (x " ++ (aux' op)  ++  " y) then (" ++ f1 ++ " y) else (" ++ f2 ++ " x))" 
							else (if (attr2==attr5) 
								then "(function x -> function y -> function z -> if (x " ++ (aux' op)  ++  " y) then (" ++ f1 ++ " z) else (" ++ f2 ++ " x))"
								else (if (attr3==attr4) 
									then "(function x -> function y -> function z -> if (x " ++ (aux' op)  ++  " y) then (" ++ f1 ++ " y) else (" ++ f2 ++ " z))" 
									else (((if (attr2==attr4) 
										 then "(function x -> function y -> function z   -> if (x " ++ (aux' op)  ++  " y) then (" ++ f1 ++ " x) else (" ++ f2 ++ " z))"
							        		 else ( if (attr3==attr5) 
											then "(function x -> function y -> function z   -> if (x " ++ (aux' op)  ++  " y) then (" ++ f1 ++ " z)  else (" ++ f2 ++ " y))"
											else "(function x -> function y -> function z -> function w -> if (x " ++ (aux' op)  ++  " y) then (" ++ f1 ++ " z) else (" ++ f2 ++ " w))"))))))),
		 nub [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4,show i5 ++ "_" ++ str5])
	     : patbinds2OCamlFunctions t varnames


-- 
patbinds2OCamlFunctions (x:t) 
	       varnames 
	= error (show x) $ patbinds2OCamlFunctions t varnames 



